//
// Copyright (c) 2003
//      Jeff Forys (jeffware@marjum.com).  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) Redistributions of
// source code must retain the above copyright notice, this list of
// conditions and the following disclaimer, (2) Redistributions in
// binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution, (3) All
// advertising materials mentioning features or use of this software
// must display the following acknowledgment: ``This product includes
// software developed by Jeff Forys (jeffware@marjum.com).'', (4)
// The name of the author may not be used to endorse or promote products
// derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//

import java.awt.*;
import java.applet.*;
import java.net.URL;
import java.net.MalformedURLException;

//
// $Id: CamClient.java,v 2.6 2005/09/22 04:53:57 forys Exp $
//
// CamClient is an applet that continually retrieves images from a given
// URL and displays them at a user-selectable interval.  The user may
// choose to freeze the current image (suspending image retrieval).
// Additionally, depending on the server-side configuration, CamClient
// allows the user to select from any number of images.
//
// Using BorderLayout, the display looks like:
//    North: Centered "Title" [Label] or unused
//    Center: Image [class CamCanvas, encased in a Panel]
//    South: Controls [class CamControl (Checkbox, Choice, etc.)]
//    East: unused
//    West: unused
//
// CamClient accepts the following parameters, and is backward compatible
// with the original "JavaCam" applet:
//    URL - Url to default image (or, URL to image files if using "images").
//    Images - colon-separated name/value pairs of image files at URL.
//    Interval - Update interval in seconds (floating point value).
//    Title - Title bar for applet.
//    Background - applet background color.
//    ControlBackground - control panel background color.
//    ControlForeground - control panel foreground color.
//    Foreground - applet foreground color (e.g. "12ff80").
//    Trace - enable trace messages in the browsers' "Java Console" window.
//    NoQuery - do not append cache-thwarting unique "?query" string on reqs.
//
// Example HTML to launch a CamClient applet:
//    <APPLET code="CamClient.class" width="100%" height="100%">
//    <PARAM name="title" value="Welcome to MY camera!">
//    <PARAM name="url" value="http://www.do.main/~webcam/pc320.jpg">
//    <PARAM name="images" value="pond-small:pc320.jpg:pond-large:pc640.jpg">
//    <PARAM name="interval" value="1">
//    </APPLET>
//
// N.B. This applet was designed to compile with JDK 1.0.2 for maximum
// browser compatibility!
//

final public class CamClient extends Applet implements Runnable
{
	static String		version = "CamClient 2.4 20030808";
	static String		copyright =
	"Copyright (c) 2003 Jeff Forys.  All rights reserved.\n\n" +
	"Redistribution and use in source and binary forms, with or without\n" +
	"modification, is permitted under the terms specified in the source\n" +
	"code, available from \"http://freshmeat.net/projects/cammgr\"";

	private static boolean	traceFlag = false;
	private static boolean	noqueryFlag = false;
	private boolean		initOkay = false;
	private Panel		imagePanel;
	private CamCanvas	imageCanvas;
	private CamControl	controlPanel = new CamControl();
	private Thread		animatorThread;
	private static long	ival = -1;
	private int		camIdx;
	private String		titleString;

	public String getAppletInfo() {
		return version;
	}

	public static void main(String args[]) {
		showVersionCopyright();
	}

	public static void showVersionCopyright() {
		System.out.println(version + "\n\n" + copyright);
	}

	// trace(message): if tracing enabled, display message System.out.
	public static void trace(String message) {
		if (traceFlag)
			System.out.println(message);
	}

	public static String makeUniqURL(String urlString) {
		if (noqueryFlag)
			return urlString;
		else
			return urlString + "?" + System.currentTimeMillis();
	}

	// sleepMilli(milliSecs): like sleep(), but ignore exceptions.
	public static void sleepMilli(long milliSecs) {
		trace("sleepMilli(" + milliSecs + "ms)...");
		try {
			Thread.currentThread().sleep(milliSecs);
		} catch (Exception ignoreEx) {
		}
	}

	// fatalError(errorMsg): fatal error, display errorMsg and die.
	public void fatalError(String errorMsg) {
		System.out.println(errorMsg);
		showStatus(errorMsg);
		sleepMilli(5000);		// pause 5 seconds...
		System.exit(1);			// may or may not work!
	}

	// setIval(ivalString): set (after validation) and return 'ival'.
	public static long setIval(String ivalString) {
		long	ivalTmp;
		try {
			ivalTmp = (long) (Float.valueOf(ivalString).
				floatValue()*1000F);
		} catch (Exception e) {
			return ival;
		}

		if (ivalTmp < 0L)
			return ival;

		return ival = ivalTmp;
	}

	// init(): parse args, configure applet layout, retrieve first image
	public void init() {
		Color	appFg = new Color(0, 102, 153);
		Color	appBg = new Color (153, 204, 153);
		Color	ctrlFg = new Color(0, 255, 0);
		Color	ctrlBg = appFg;
		String	urlString;

		showVersionCopyright();

		if (initOkay) {
			trace("init(): already initialized: skipping...");
			return;
		}

		// OPT ARG "Trace": if supplied, enable debug messages.
		if (getParameter("Trace") != null)
			controlPanel.setTrace(traceFlag = true);

		trace(">>> init()");

		// ARG "URL": image URL (or return with an error message).
		if ((urlString = getParameter("URL")) == null) {
			fatalError("CamClient: CfgError: no URL parameter");
			return;
		}

		// ARG "Interval": update interval in secs (default 5).
		if (setIval(getParameter("Interval")) < 0)
			ival = 5 * 1000;

		// OPT ARG "NoQuery": If supplied, do not append a unique
		// query string to each request.  Enabling this option
		// can create image caching problems on some browsers.
		// You probably do NOT want to set this option...
		if (getParameter("NoQuery") != null)
			noqueryFlag = true;

		trace("init(): arg [Trace=" + traceFlag +
			"]\ninit(): arg [URL=" + urlString +
			"]\ninit(): arg [Interval=" + ival +
			"]\ninit(): arg [NoQuery=" + noqueryFlag + "]");

		// OPT ARG "Title": Title for "north" border.
		titleString = getParameter("Title");

		// OPT ARG "Foreground": applet foreground color.
		// OPT ARG "Background": applet background color.
		// OPT ARG "ControlForeground": control panel foreground color.
		// OPT ARG "ControlBackground": control panel background color.
		try {
			String cstr;
			if ((cstr = getParameter("Foreground")) != null)
				appBg = new Color(Integer.parseInt(cstr, 16));
			if ((cstr = getParameter("Background")) != null)
				appBg = new Color(Integer.parseInt(cstr, 16));
			if ((cstr = getParameter("ControlForeground")) != null)
				ctrlFg = new Color(Integer.parseInt(cstr, 16));
			if ((cstr = getParameter("ControlBackground")) != null)
				ctrlBg = new Color(Integer.parseInt(cstr, 16));
		} catch (Exception ignoreEx) {
			fatalError("CamClient: CfgError: invalid color format");
		}

		// OPT ARG "Images": optional list of image:filename pairs
		if (controlPanel.addCameras(urlString,
				getParameter("Images")) == false) {
			fatalError("CamClient: CfgError: Invalid URL");
			return;
		}

		// Configure the BorderLayout...
		setLayout(new BorderLayout());
		setForeground(appFg);
		setBackground(appBg);
		setFont(new Font("Helvetica", Font.PLAIN, 14));

		// North gets the title, iff we have one.
		if (titleString != null) {
			Label titleLabel = new Label(titleString,Label.CENTER);
			titleLabel.setFont(new Font("Helvetica",Font.BOLD,30));
			add("North", titleLabel);
		}

		// Center gets the image (in a Panel for manageability).
		imagePanel = new Panel();
		add("Center", imagePanel);
		imageCanvas = new CamCanvas(		// pull first image
			controlPanel.currentCameraUrl(), this);
		imagePanel.add(imageCanvas);

		// South gets the control panel.
		camIdx = controlPanel.setValues(imageCanvas.getWidth(),
			Float.toString(((float)ival) / 1000F), ctrlFg, ctrlBg);
		add("South", controlPanel);

		validate();
		repaint();

		initOkay = true;
		trace("<<< init()");
	}

	// Start the animator thread, which will call run().
	public void start() {
		trace("CamClient.start()");
		if (animatorThread == null)
			animatorThread = new Thread(this);
		animatorThread.start();
	}

	// Stop the animator thread (if currently running).
	public void stop() {
		trace("CamClient.stop()");
		if (animatorThread != null) {
			animatorThread.stop();
			animatorThread = null;
		}
		super.stop();
	}

	// Loop pulling/displaying images; check controls on each iteration.
	public void run() {
		Thread myThread = Thread.currentThread();
		long nextTimeMs, sleepTimeMs;

		trace(">>> run()");

		myThread.setPriority(Thread.MIN_PRIORITY);	// low priority

		// Enter the main image fetching loop, where 'nextTimeMs'
		// is the next time we will pull an image.
		nextTimeMs = System.currentTimeMillis();
		while (animatorThread == myThread) {
			// Check these options before pulling a new image.
			traceFlag = controlPanel.isTrace();	// opt trace
			if (controlPanel.isFreeze()) {		// opt freeze
				sleepMilli(1000);
				nextTimeMs = System.currentTimeMillis();
				continue;
			}

			// If user has chosen a new camera, remove the old
			// imageCanvas, get new image/Canvas and repaint.
			// Otherwise, just grab image from current camera.
			if (camIdx != controlPanel.getCamIdx()) {
				imagePanel.remove(imageCanvas);
				imageCanvas = new CamCanvas(
					controlPanel.currentCameraUrl(), this);
				imagePanel.add(imageCanvas);
				camIdx = controlPanel.setValues(
					imageCanvas.getWidth());
				validate();
				repaint();
			} else
				imageCanvas.newImage();

			// Calculate time of next pull relative to this one
			// then sleep for the duration.  A minimal sleep of
			// 100ms is enforced (i.e., 10 frames/second).
			nextTimeMs += ival;
			sleepTimeMs = nextTimeMs - System.currentTimeMillis();

			// For perfect timing, 'nextTimeMs' is incremented
			// relative to 'ival'.  However, if the browser stops
			// or freezes for any period of time, this can result
			// in a bunch of 100ms requests to "catch up."  To
			// resolve this, if 'sleepTimeMs' is negative, reset
			// 'nextTimeMs' to the current time.
			if (sleepTimeMs < 0)
				nextTimeMs = System.currentTimeMillis();
			if (sleepTimeMs <= 100)
				sleepTimeMs = 100;
			sleepMilli(sleepTimeMs);
		}
	}
}

//
// CamCanvas manages an image using the java.awt.MediaTracker class.
// Images are grabbed from the URL supplied in the constructor (i.e.,
// an instantiation of a CamCanvas always grabs the same image).
//
// N.B. The constructor also requires the controlling applet to access
// its java.applet.Applet.getImage() method; using Toolkit.getImage()
// can cause security exceptions on some browsers.
//
class CamCanvas extends Canvas {
	private String		urlString;
	private Applet		applet;
	private MediaTracker	mt;
	private Image		image = null;
	private int		width, height;

	public CamCanvas(String urlString, Applet applet) {
		super();

		this.urlString = urlString;
		this.applet = applet;
		newImage();
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}

	public Dimension getMinimumSize() {
		return new Dimension(getWidth(), getHeight());
	}

	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	// newImage(): pull a new image and draw it to the canvas.
	public synchronized void newImage() {
		URL url;

		if (image != null)		// flush cache to force update
			image.flush();

		try {
			url = new URL(CamClient.makeUniqURL(urlString));
		} catch(MalformedURLException ex) {
			//
			// This should never happen; the URL was already
			// tested for validity and, at most, all that
			// makeUniqURL() did was append "?" + "<digits>".
			//
			ex.printStackTrace();
			return;
		}

		// If we have an applet reference, use it to overcome
		// security exceptions w/Toolkit getImage() in an applet.
		if (applet == null)
			image = Toolkit.getDefaultToolkit().getImage(url);
		else
			image = applet.getImage(url);

		mt = new MediaTracker(this);
		mt.addImage(image, 1);
		try {
			mt.waitForID(1);
		} catch(Exception e) {
			e.printStackTrace();
			return;
		}

		// Now that we have the image, (re)set WxD accordingly.
		width = image.getWidth(this) > 0? image.getWidth(this): 320;
		height = image.getHeight(this) > 0? image.getHeight(this): 240;

		repaint();
	}

	// update(): avoid clear (no flicker); assume images are same size
	public void update (Graphics g) {
		paint (g);
	}

	// paint(): draw the image.
	public synchronized void paint(Graphics g) {
		CamClient.trace("CamCanvas.paint(): "+width+"x"+height);
		if  ((mt.statusID(1, false) & MediaTracker.ERRORED) != 0 ||
				image == null) {
			g.setColor(Color.red);
			g.fillRect(0, 0, width, height);
			return;
		}
		g.drawImage(image, 0, 0, width, height, this);
	}
}

//
// CamControl manages all the user controls (Checkboxes, Choices, etc).
// When instantiated, it sets up all of these as appropriate.  It
// also installs an event handler to manage this control panel.
//
// The camera URL strings are maintained in cameraUrls[], where the
// index into the array corresponds to the optCameraChoice[] Choice.
// The number of cameras is quietly limited to 32 -- a reasonable number,
// given that the selection is made through a "Choice" -- and it's also
// less expensive that instantiating a Vector.
//
// After invoking the constructor, the caller is expected to call back
// through setValues() with several parameters available after an image
// is retrieved.
//
class CamControl extends Panel {
	private Checkbox	optTraceBox, optFreezeBox;
	private TextField	optIvalText;
	private Choice		optCameraChoice;
	final int		MAXCAMERA = 32;
	final int		height = 50;
	private int		width = 320;
	private String		cameraUrls[];		// configured URLs
	private int		camCnt = 0;		// camera count

	public CamControl() {
		super();

		cameraUrls = new String[MAXCAMERA];
		optCameraChoice = new Choice();
		add(optCameraChoice);

		optFreezeBox = new Checkbox();
		optFreezeBox.setLabel("Freeze");
		add(optFreezeBox);

		optTraceBox = new Checkbox();
		optTraceBox.setLabel("Trace");
		add(optTraceBox);

		Label ivalLabel = new Label("Refresh:", Label.RIGHT);
		add(ivalLabel);
		optIvalText = new TextField("1.0", 3);
		add(optIvalText);
	}

	public Dimension getMinimumSize() {
		return new Dimension(width, height);
	}

	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	// addCameras(): Add all camera URL strings to cameraUrls[] array.
	// If 'cameraList' is null, only the "default" camera is available.
	// Return true on success, false on error.  Errors here are fatal.
	public boolean addCameras(String urlString, String cameraList) {
		int urlLen = urlString.length();
		String camFilDef = null;

		// If supplied, parse the "Images" parameter list found in
		// 'cameraList' and construct URLs using 'urlString' as base.
		if (urlLen > 0 && cameraList != null) {
			// Locate file name in the URL, fail no slash found.
			int camFilIdx = urlLen - 1;
			for ( ; camFilIdx >= 0; camFilIdx--) {
				char slash = urlString.charAt(camFilIdx);
				if (slash == '/' || slash == '\\')
					break;
			}
			if (camFilIdx++ < 0)		// invalid URL
				return false;

			// It's okay if the URL only specifies the directory,
			// however for backward compatibility, URL may also
			// supply the file name... remember it as "default."
			if (camFilIdx < urlLen)	
				camFilDef = urlString.substring(camFilIdx);

			// Walk the colon-separated Name:File pairs and
			// construct the optCameraChoice selections and
			// corresponding cameraUrls[] array.
			String camNam = null, camUrl;
			int camLen = cameraList.length();
			for (int c, cp = 0; cp < camLen; cp = c + 1) {
				// leading or double colon is a format error.
				if ((c = cameraList.indexOf(':', cp)) == cp) {
					CamClient.trace("Images: format error");
					return false;
				}
				if (c < 0)			// last one
					c = camLen;

				// First comes the camera name.
				if (camNam == null) {
					camNam = cameraList.substring(cp, c);
					continue;
				}

				// Second is the file name.  If we find
				// "default" file in list, set 'camFilDef'
				// to null again (since we dont need it).
				String camFil = cameraList.substring(cp, c);
				if (camFilDef != null &&
				    camFilDef.compareTo(camFil) == 0)
					camFilDef = null;
				camUrl = urlString.substring(0, camFilIdx) +
					camFil;
				CamClient.trace("addCameras() <" + camNam +
					"=" + camUrl + ">");

				// Add camera to the choice menu and array.
				if (camCnt < MAXCAMERA) {
					try {	// check that URL is valid
						URL tmpUrl = new URL(camUrl);
					} catch(MalformedURLException ex) {
						return false;
					}
					optCameraChoice.addItem(camNam);
					cameraUrls[camCnt++] = camUrl;
				}

				camNam = null;
			}
			if (camNam != null) {		// still have a name!
				CamClient.trace("Images: missing image file");
				return false;
			}
		}

		// Add the default camera to the choice menu and array.
		// Select "default" as initial choice, or first in menu.
		if ((cameraList == null || camFilDef != null) &&
		    camCnt < MAXCAMERA) {
			CamClient.trace("addCameras() <default=" +
				urlString + ">");
			try {	// check that URL is valid
				URL tmpUrl = new URL(urlString);
			} catch(MalformedURLException ex) {
				return false;
			}
			optCameraChoice.addItem("default");
			cameraUrls[camCnt++] = urlString;
			optCameraChoice.select(camCnt - 1);
		} else
			optCameraChoice.select(0);

		return true;
	}

	// getCamIdx(): Return index of currently selected camera.
	public int getCamIdx() {
		return optCameraChoice.getSelectedIndex();
	}

	// currentCameraUrl(): Return URL corresponding to the current camera.
	public String currentCameraUrl() {
		return cameraUrls[getCamIdx()];
	}

	// isFreeze(): Return true is optFreezeBox is selected, false o/w.
	public boolean isFreeze() {
		return optFreezeBox.getState();
	}

	// isTrace(): Return true is optTraceBox is selected, false o/w.
	public boolean isTrace() {
		return optTraceBox.getState();
	}

	// setTrace(): set optTraceBox to the specified state.
	public void setTrace(boolean trace) {
		optTraceBox.setState(trace);
	}

	// setValues(): resize image, return current camera.
	public int setValues(int width) {
		this.width = width;
		resize(getPreferredSize());
		return getCamIdx();
	}

	// setValues(): resize image, set colors, and return current camera.
	public int setValues(int width, String ivalString, Color fg, Color bg) {
		optIvalText.setText(ivalString);
		setForeground(fg);
		setBackground(bg);
		return setValues(width);
	}

	// paint(): draw the control panel.
	public synchronized void paint(Graphics g) {
		CamClient.trace("CamControl.paint(): "+width+"x"+height);
		super.paint(g);
	}

	// action(): control panel event handler.
	public synchronized boolean action(Event e, Object arg) {
		//
		// An earlier version started/stopped/suspended threads.
		// While this made for "instant" updates (e.g., CamClient
		// was immediately restarted with the new configuration),
		// it also induced problems in ancient browsers.
		//
		// This implementation simply sets the local config vars,
		// where it's the responsibility of CamClient to check and
		// reconfigure itself as necessary.
		//
		if (e.target == optFreezeBox) {
			CamClient.trace("action: optFreezeBox event");
		} else if (e.target == optTraceBox) {
			CamClient.trace("action: optTraceBox event");
		} else if (e.target == optIvalText) {
			CamClient.trace("action: optIvalText event");
			optIvalText.setText(Float.toString( ((float)
				CamClient.setIval(optIvalText.getText()) )
				/ 1000F));
		} else if (e.target == optCameraChoice) {
			CamClient.trace("action: optCameraChoice event");
		} else {
			return false;
		}
		return true;
	}
}
