/*
** Copyright (c) 2002, 2003, 2005
**      Jeff Forys (jeffware@marjum.com).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeffware@marjum.com).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

const char *Version = "1.5 20051120";

const char *CopyrightVersion = "cammgr %s%s%s\n\nCopyright (c) 2002, 2003, 2005 Jeff Forys.  All rights reserved.\n\nRedistribution and use in source and binary forms, with or without\nmodification, is permitted under the terms specified in the source\ncode (and manual pages).\n";

#ifndef lint
static char rcsid[] = "$Id: cammgr.c,v 1.52 2016/02/07 01:50:26 forys Exp $";
#endif

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <pwd.h>
#include <dirent.h>
#include <limits.h>

#include "config.h"
#include "parser.h"
#include "compat.h"

/*
 * Various enums (and associated state messages when applicable).
 */
typedef	enum	{ act_none, act_up, act_cap, act_autocap, act_down } action_t;
char *actmsg[]=	{ "NONE", "UP", "CAPTURE", "AUTOCAPTURE", "DOWN" };

typedef enum	{ mnt_on,   mnt_off,   mnt_offloop    } optmnt_t;
char *mntmsg[]=	{ "MOUNT",  "UNMOUNT", "UNMOUNT_LOOP" };

typedef	enum	{ cli_add, cli_prune, cli_clrcam, cli_clrall, cli_rsamp }
								clicmd_t;

typedef	enum	{ cap_start, cap_nsamp, cap_stop, cap_closeall } capcmd_t;

typedef	enum	{ pid_req, pid_cur, pid_curonly } pidtyp_t;

typedef	int	camidx_t;

/*
 * Globals.
 */
static struct confdata_s *Cf;		/* configuration data */
volatile action_t CurState[CONFCAMMAX];	/* up, autocap, capture, or down */
volatile pid_t CapPid[CONFCAMMAX];	/* capture process pids */
uid_t MyUid;				/* callers original real uid */
int TotCam = 0;				/* total configured cameras */
int LogFd = -1;				/* event-driven log file descriptor */

volatile pid_t capman_pid;		/* pid for capture child proc */
volatile int capman_stat;		/* exit status for capture child proc */

/*
 * Signals are our friends.
 *
 * SIGRESET()       - reset handlers and unblock signals (used post-fork()).
 * sigmask_openpid  - signals harmful during openpid() (i.e., SIGTSTP).
 * sigmask_newstate - signals that cause state transitions.
 */
#define	SIGRESET()						\
do {								\
	struct sigaction _sigact;				\
	sigset_t _sigmask;					\
								\
	_sigact.sa_flags = 0;					\
	_sigact.sa_handler = SIG_DFL;				\
	sigaction(SIGHUP, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGINT, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGQUIT, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGPIPE, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGTERM, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGTSTP, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGCHLD, &_sigact, (struct sigaction *)NULL);	\
	sigaction(SIGUSR1, &_sigact, (struct sigaction *)NULL);	\
	sigemptyset(&_sigmask);					\
	sigprocmask(SIG_SETMASK, &_sigmask, NULL);		\
} while (0)

sigset_t sigmask_openpid;
sigset_t sigmask_newstate;

/*
 * Close event-driven log file (if open) and reset LogFd to -1
 */
#define	CLOSE_LOGFD()						\
do {								\
	if (LogFd != -1)					\
		(void) close(LogFd);				\
	LogFd = -1;						\
} while (0)


/*
 * Prototypes!
 */
int	initcam(void);
void	actor(void);
void	newstate(action_t *);
void	acthdlr(int), reaper(int), caphdlr(int);
int	strmatch(const char *, const char *);
int	usage(const char *, const char *);
int	version(void);
void	openpid(void);
pid_t	readpid(pidtyp_t, action_t *, uid_t *);
void	writepid(pidtyp_t, const action_t *, pid_t);
void	closepid(void);
pid_t	mntexec(optmnt_t op);
pid_t	capexec(capcmd_t, camidx_t, unsigned int);
pid_t	cliexec(const char *, const char *, const char *, const char *);
void	writedefimg(struct confcam_s *);
int	climgr(clicmd_t, camidx_t, const char *, const char *, action_t *);
int	parselog(char *, int, action_t *);
action_t camshow(const char *);
void	clishow(void);

/*
 * cammgr - manage a web camera.
 *
 * cammgr [-fvw] [ up|capture|down [ <camera1> ... <cameraN> ] ]
 */
int
main(int argc, char *argv[])
{
	extern int optind;
	action_t req, camreq[CONFCAMMAX];
	struct sigaction act;
	struct passwd *pw;
	struct confcam_s *cfc;
	pid_t pid;
	const char *errmsg;
	char *progname;
	int i, fd, opt_force = 0, opt_wait = 0;

	SETPROGNAME(progname, argv);

	MyUid = getuid();

	/* SIGTSTP blocked while pid file is opened (no sleeping w/lock) */
	sigemptyset(&sigmask_openpid);
	sigaddset(&sigmask_openpid, SIGTSTP);

	/* SIGCHLD is necessary or syscalls can return EINTR */
	sigemptyset(&sigmask_newstate);
	sigaddset(&sigmask_newstate, SIGTERM);
	sigaddset(&sigmask_newstate, SIGCHLD);
	sigaddset(&sigmask_newstate, SIGUSR1);

	/* set umask for owner rw, group and other read-only */
	(void) umask(S_IWGRP|S_IWOTH);

	/* parse flags, return usage() error if not defined */
	while ((i = getopt(argc, argv, "fvw")) != -1)
		switch (i) {
		    case 'f':
			opt_force = 1;
			break;
		    case 'v':
			return version();
		    case 'w':
			opt_wait = 1;
			break;
		    case '?':
		    default:
			return usage(progname, "unknown flag");
		}
	argc -= optind;
	argv += optind;

	/* open log file to capture any configuration file errors */
	openlog(progname, LOG_PERROR, SYSLOGFAC);

	if ((Cf = confread(CFGFIL, &errmsg)) == NULL) {
		syslog(LOG_ERR, "%s: %s", CFGFIL, errmsg);
		return 1;
	}
	TotCam = Cf->camera_e - Cf->camera;

	/* ensure user has permission to run application */
	if ((pw = getpwuid(MyUid)) == NULL) {
		syslog(LOG_ERR, "getpwuid: no such user: %d", MyUid);
	} else if (MyUid != 0 && strstr(Cf->camusers, ":*:") == NULL) {
		char usrnam[128 + 4];		/* max user name */
		snprintf(usrnam, sizeof usrnam, ":%s:", pw->pw_name);
		if (strstr(Cf->camusers, usrnam) == NULL) {
			errno = EPERM;
			syslog(LOG_ERR, "user %s: %m", pw->pw_name);
			pw = NULL;		/* force usage error */
		}
	}
	if (pw == NULL)
		return usage(progname, "permission denied");

	/* dont need saved uid: effective -> real (also for readpid/kill) */
#ifdef	USE_SETREUID
	setregid(getegid(), -1);
	setreuid(geteuid(), -1);
#else
	setgid(getegid());
	setuid(geteuid());
#endif

	if (argc == 0) {			/* request status */
		if (camshow(progname) == act_down)
			return 3;
		clishow();
		return 0;
	} else if (strmatch("up", *argv))	/* request up */
		req = act_up;
	else if (strmatch("capture", *argv))	/* request capture */
		req = act_cap;
	else if (strmatch("down", *argv))	/* request down */
		req = act_down;
	else
		return usage(progname, "invalid action specified");

	for (i = 0; i < CONFCAMMAX; i++)	/* initialize per-camera reqs */
		camreq[i] = (argc == 1)? req: act_none;

	/* parse rest of command line for camera names */
	for (argv++, argc--; argc > 0; argv++, argc--) {
		for (cfc = Cf->camera, i = 0; cfc < Cf->camera_e; cfc++, i++)
			if (strcmp(*argv, cfc->name) == 0) {
				camreq[i] = req;
				break;
			}
		if (cfc == Cf->camera_e) {
			fprintf(stderr, "%s: unknown camera: %s\n",
				progname, *argv);
			return 1;
		}
	}

	/* block SIGTSTP, then get current state of server (and its pid) */
	sigprocmask(SIG_BLOCK, &sigmask_openpid, NULL);
	openpid();
	pid = readpid(pid_cur, NULL, NULL);

	/* if current state is requested state, there's nothing to do */
	if (!opt_force) {
		for (i = 0; i < TotCam; i++)
			if (camreq[i] != CurState[i])
				break;
		if (i == TotCam) {
			closepid();
			sigprocmask(SIG_UNBLOCK, &sigmask_openpid, NULL);
			if (opt_wait)
				(void) camshow(progname);
			return 0;
		}
	}

	/*
	 * If no server is running, start one to handle the transition.
	 * Otherwise, write request and signal the already-running server.
	 */
	if (pid != 0) {
		writepid(pid_req, camreq, pid);
		closepid();
		if (kill(pid, SIGUSR1) < 0) {
			syslog(LOG_ERR, "server communication failure (%d): %m",
				pid);
			return 2;
		}
		sigprocmask(SIG_UNBLOCK, &sigmask_openpid, NULL);
	} else {
		/*
		 * If currently "down" and request was for something
		 * other than "down", fork and become the manager.
		 */
		switch (pid = fork()) {
		    case -1:		/* error */
			syslog(LOG_ERR, "fork: %m");
			closepid();
			sigprocmask(SIG_UNBLOCK, &sigmask_openpid, NULL);
			return 2;

		    case 0:		/* child */
			closepid();		/* parent still holds lock */
			closelog();		/* soon we'll open a new one */
#ifdef	HAS_SETPROCTITLE
			setproctitle("server");
#endif
			/*
			 * Daemonize; set up a new session, set working
			 * dir to '/' close stdin/stdout/stderr, etc.
			 */
			(void) setsid();
			(void) chdir("/");
			if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
				int maxdsiz = getdtablesize();
				(void) dup2(fd, STDIN_FILENO);
				(void) dup2(fd, STDOUT_FILENO);
				(void) dup2(fd, STDERR_FILENO);
				(void) close(fd);
				for (fd = 0; fd < maxdsiz; fd++)
					(void) close(fd);
			}

			/*
			 * Open log file, set descriptor to close on exec.
			 *
			 * Note: STDIN/STDOUT/STDERR are opened first to
		`	 * ensure that syslog does not use any of these
			 * descriptors.  After that, they are closed.
			 */
			if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
				(void) dup2(fd, STDIN_FILENO);
				(void) dup2(fd, STDOUT_FILENO);
				(void) dup2(fd, STDERR_FILENO);
			}
			openlog(progname, LOG_NDELAY | LOG_PID, SYSLOGFAC);
			fcntl(0, F_SETFD, 1);
			(void) close(STDIN_FILENO);
			(void) close(STDOUT_FILENO);
			(void) close(STDERR_FILENO);

			syslog(LOG_NOTICE, "server started (uid:%u)...", MyUid);
			MyUid = getuid();

			/* server runs relative to camera directory */
			if (chdir(Cf->camdir) < 0) {
				syslog(LOG_ERR, "chdir(%s): %m", Cf->camdir);
				_exit(2);
			}

			if (initcam())		/* do autoconf part of 'Cf' */
				_exit(2);

			/*
			 * Ignore harmful signals:
			 *	SIGHUP|SIGINT|SIGQUIT|SIGPIPE|SIGTSTP
			 * Set up signal handlers:
			 *	SIGCHLD -> reaper()
			 *	SIGTERM|SIGUSR1 -> acthdlr()
			 */
			act.sa_flags = 0;
			act.sa_handler = SIG_IGN;
			sigaction(SIGHUP, &act, (struct sigaction *)NULL);
			sigaction(SIGINT, &act, (struct sigaction *)NULL);
			sigaction(SIGQUIT, &act, (struct sigaction *)NULL);
			sigaction(SIGPIPE, &act, (struct sigaction *)NULL);
			sigaction(SIGTSTP, &act, (struct sigaction *)NULL);

			act.sa_mask = sigmask_newstate;
			act.sa_handler = reaper;
			sigaction(SIGCHLD, &act, (struct sigaction *)NULL);
			act.sa_handler = acthdlr;
			sigaction(SIGTERM, &act, (struct sigaction *)NULL);
			sigaction(SIGUSR1, &act, (struct sigaction *)NULL);

			actor();
			/*NOTREACHED*/
			_exit(2);

		    default:		/* parent */
			/* write current state and request for spawned daemon */
			writepid(pid_cur, NULL, pid);
			writepid(pid_req, camreq, pid);
			break;
		}
	}

	closepid();
	sigprocmask(SIG_UNBLOCK, &sigmask_openpid, NULL);

	/* if requested, wait for a state transition to happen before exit */
	if (opt_wait)
		(void) camshow(progname);

	return 0;
}

/*
 * initcam() - initialize auto-configuration section of Cf->camera[].
 *
 * On successful return, the auto-configuration section of each
 * camera in Cf->camera[] has been initialized.  Specifically:
 *    o 'defimg' and 'defsiz' were initialized to arrays; each
 *	element in 'defimg[]' is at least a NULL pointer and a
 *	corresponding '0' in 'defsiz[]'.  Any non-NULL pointers
 *	point to the contents of the corresponding default image
 *	file of size 'defsiz[]' (still '0' if file was empty).
 *    o If 'capcmd' is non-NULL, variable substitution has been
 *	performed on it.  If it is NULL, it remains NULL.
 *    o 'capfun' points to an internal function or NULL.
 */
int
initcam(void)
{
	const char *multifil = "Warning: %s: image file multiply defined: %s";
	struct confcam_s *cfc;
	struct stat sb;
	char fil[PATH_MAX], *filnam;
	int i, remcnt, imgcnt;

	/*
	 * Initialize 'fil' to path to 'Cf->defimgdir'.
	 * Position 'filnam' to byte after trailing slash.
	 * Ultimately set 'remcnt' to space remaining in 'fil'.
	 */
	if (Cf->defimgdir != NULL) {
		snprintf(fil, PATH_MAX, "%s/", Cf->defimgdir);
		remcnt = strlen(fil);
		filnam = fil + remcnt;
		remcnt = PATH_MAX-1 - remcnt;
	} else {
		remcnt = PATH_MAX - 1;		/* keep GCC quiet */
		filnam = fil;
	}

	for (cfc = Cf->camera; cfc < Cf->camera_e; cfc++) {
		char *ccp, *occp, *ncapcmd;

		/*
		 * If compiled with CAPFUN, it must contain the name of
		 * a function that'll handle the capture operation for
		 * all cameras (code can multiplex based on cfc->name).
		 */
#ifdef	CAPFUN
		{
			extern void (*CAPFUN)(const struct confcam_s *);
			cfc->capfun = CAPFUN;
		}
#else
		cfc->capfun = NULL;
#endif

		/* calculate maximum number of images for this camera */
		imgcnt = cfc->imgfil_e - cfc->imgfil;

		if ((occp = cfc->capcmd) != NULL) {
			/*
			 * Determine maximum possible number of '%' escapes,
			 * then allocate enough memory for expanded command.
			 * No config file line is bigger than CONFLINELEN,
			 * no image file name is bigger than CONFIMGLEN.
			 */
			for (occp = cfc->capcmd, i = 0;
						(ccp=strchr(occp, '%')) != NULL;
						occp = ccp + 1, i++)
				;
			ncapcmd = malloc(CONFLINELEN + CONFIMGLEN * i);
			if (ncapcmd == NULL) {
nomem:
				syslog(LOG_ERR, "capcmd: %s: out of memory",
					cfc->name);
				return 1;
			}
			for (occp = cfc->capcmd, ccp = ncapcmd; *occp != '\0';
						ccp++, occp++) {
				i = -1;
				if ((*ccp = *occp) == '%') {
					/*
					 * First (copied) char was escape '%'.
					 * If another follows ("%%"), ignore
					 * it as we already condensed to '%'.
					 * If a digit follows, convert number
					 * to a positional parameter.
					 */
					if (*(++occp) == '%')
						continue;
					else if (isdigit(*occp))
						i = strtol(occp, &occp, 10);
					occp--;		/* will reincrement */
				}
				if (i >= 0 && i < imgcnt) {
					const char *l = cfc->imgfil[i];
					*ccp++ = '_';	/* overrite '%' */
					while (*l != '\0')
						*ccp++ = *l++;
					ccp--;		/* will reincrement */
				}
			}
			*ccp = '\0';

			free(cfc->capcmd);		/* out with the old */
			cfc->capcmd = ncapcmd;		/* in with the new */
		}

		/* warn if same image file is used for multiple cameras */
		for (i = 0; i < imgcnt; i++) {
			struct confcam_s *tcfc;
			int dupcnt = 0;
			char **p;
			for (tcfc = Cf->camera; tcfc < Cf->camera_e; tcfc++)
				for (p = tcfc->imgfil; p < tcfc->imgfil_e; p++)
					if (strcmp(cfc->imgfil[i], *p) == 0 &&
							cfc != tcfc->ovride &&
							tcfc != cfc->ovride)
						dupcnt++;
			if (dupcnt > 1)
				syslog(LOG_WARNING, multifil, cfc->name,
					cfc->imgfil[i]);
		}

		/* note: using calloc() to initialize with NULL/zero's */
		if ((cfc->defimg = calloc(imgcnt, sizeof(char *))) == NULL ||
		    (cfc->defsiz = calloc(imgcnt, sizeof(size_t))) == NULL) {
			if (cfc->defimg != NULL) {
				free(cfc->defimg);
				cfc->defimg = NULL;
			}
			goto nomem;
		}

		if (Cf->defimgdir == NULL)	/* no default image dir */
			continue;

		for (i = 0; i < imgcnt; i++) {	/* read in default images */
			FILE *fp;
			char *errmsg = NULL;
			ssize_t rret;
			size_t siz;

			/* append file name to path */
			strncpy(filnam, cfc->imgfil[i], remcnt);
			if (stat(fil, &sb) < 0 ||
					(sb.st_mode & S_IFMT) != S_IFREG) {
				errmsg = "stat failed";
				goto nodefimgok;
			}
			/* cast makes assumption that image sizes are < 2GB */
			siz = cfc->defsiz[i] = (size_t)sb.st_size;
			/* empty file (siz == 0) ok, cant be NULL ptr tho */
			if ((cfc->defimg[i] = malloc(siz? siz: 1)) == NULL) {
				errmsg = "out of memory";
				goto nodefimg;
			}
			if ((fp = fopen(fil, "r")) == NULL) {
				errmsg = "open failed";
				goto nodefimg;
			}
			rret = fread(cfc->defimg[i], sizeof(char), siz, fp);
			fclose(fp);
			if (rret == siz)		/* got it! */
				continue;

			errmsg = (rret > 0)? "read incorrect size":
				"read failed";
nodefimg:
			syslog(LOG_WARNING, "%s: no default image: %s: %s",
				cfc->name, fil, errmsg);
nodefimgok:
			if (cfc->defimg[i] != NULL)
				free(cfc->defimg[i]);
			cfc->defimg[i] = NULL;
			cfc->defsiz[i] = 0;
		}
	}

	return 0;
}

/* CAM_REQTYP - set 'selcam' TRUE if 'camlst[]' targets select cameras */
#define	CAM_REQTYP(selcam, camlst)				\
	for (selcam = 0; selcam < TotCam; selcam++)		\
		if (camlst[selcam] != act_none)			\
			break;					\
	selcam = (selcam == TotCam)? 0: 1

/*
 * actor()
 *
 * This is the main, event-driven loop.  There are two event classes:
 * signals and log file.  "Signal driven" is always supported while
 * "Log driven" is only supported if we are able to open Cf->logfil.
 *
 * This function never returns.
 */
void
actor(void)
{
	const char *logdrvmsg = "log-driven mode";
	struct timeval tval;
	fd_set ibits;
	action_t req, camreq[CONFCAMMAX];
	int n, opt_cam, dispmsg = -1, snooze = 0;
	char buf[8192];

	req = act_up;	/* anything but 'act_none' will do here */

	for(;;) {
		/* if not 'act_none', someone is requesting a state change */
		if (req != act_none) {
			sigprocmask(SIG_BLOCK, &sigmask_newstate, NULL);
			newstate(req == act_up? NULL: camreq);
			sigprocmask(SIG_UNBLOCK, &sigmask_newstate, NULL);
			req = act_none;
		}

		/* if no log file defined, we'll always be signal-driven... */
		if (Cf->logfil == NULL) {
			sigset_t emptymask;
			sigemptyset(&emptymask);
			sigsuspend(&emptymask);
			continue;
		}

		/*
		 * Ask climgr() to do some pruning since we may have
		 * come here on a read/select error (e.g., closed
		 * file/fifo) while there were outstanding clients.
		 */
		n = climgr(cli_prune, (camidx_t)0, NULL, NULL, camreq);
		CAM_REQTYP(opt_cam, camreq);
		if (opt_cam) {
			req = act_autocap;	/* handle the state changes */
			continue;
		}

		/*
		 * A couple things can happen that could send us into
		 * a CPU-sucking loop.  For example, trying to open a
		 * file that doesnt exist, or opening a fifo with no
		 * reader makes read() return EOF.  In all cases,
		 * 'snooze' was set to sleep a bit rather than churn.
		 */
		if (snooze) {
			sleep((Cf->usertout / 4) + 1);
			snooze = 0;
		}

		/*
		 * If the log file isnt open, try (again) to open it.
		 * climgr() returned the number of outstanding clients.
		 * If none, block on the open; if a signal arrives,
		 * open will return -1/EINTR and we loop thru again.
		 *
		 * N.B. If climgr() reports no clients, it is correct
		 * to block.  It's impossible to acquire clients due
		 * to a signal; only a read from the log file (which
		 * we are currently trying to open) can get a client.
		 */
		if (LogFd < 0) {		/* signal driven */
			if ((LogFd = open(Cf->logfil,
					((n > 0)? O_NONBLOCK: 0) | O_RDONLY,
					0600)) < 0 && errno != EINTR) {
				if (dispmsg < 1) {
					syslog(LOG_NOTICE,
						"%s disabled: open(%s): %m",
						logdrvmsg, Cf->logfil);
					dispmsg = 1;
				}
				snooze = 1;
			}
			if (LogFd >= 0) {	/* if reg file, start at end */
				lseek(LogFd, (off_t)0, SEEK_END);
				fcntl(LogFd, F_SETFD, 1);   /* close on exec */
			}
			continue;
		}

		if (dispmsg == -1 || dispmsg == 1) {
			syslog(LOG_NOTICE, "%s enabled: %s", logdrvmsg,
				Cf->logfil);
			dispmsg = 0;
		}

		/* signal + log driven */
		FD_ZERO(&ibits);
		FD_SET(LogFd, &ibits);
		tval.tv_sec = (Cf->usertout / 4) + 1;
		tval.tv_usec = 0;
		switch (n = select(LogFd+1, &ibits, NULL, NULL,
						(n > 0)? &tval: NULL)) {
		    case -1:		/* select() error */
			/*
			 * If interrupted by signal, probably had a
			 * state change, o/w give up on this log file.
			 */
			if (errno == EINTR)
				continue;
			CLOSE_LOGFD();
			snooze = 1;
			break;

		    case 0:		/* inactivity timeout */
			break;

		    default:		/* more log data to parse */
			switch(n = read(LogFd, buf, sizeof buf)) {
			    case -1:
				if (errno == EINTR)
					continue;
				CLOSE_LOGFD();
				break;
			    case 0:
				snooze = 1;
				break;
			    default:
				if (parselog(buf, n, camreq)) {
					CAM_REQTYP(opt_cam, camreq);
					if (opt_cam)
						req = act_autocap;
				}
				break;
			}
			break;
		}
	}
	/*NOTREACHED*/
}
#undef	CAM_REQTYP

/*
 * newstate(camreqlog) - handle state transitions for list of cameras.
 *
 * Open/lock PID file through readpid(), then carry out the state
 * transition(s).  If 'camreqlog' is NULL, read the request from
 * the PID file (initial or signal-driven request).  If 'camreqlog'
 * is "1", we are fielding a SIGTERM and must shut down.  Otherwise,
 * the request is in 'camreqlog' (log-driven).
 *
 * Caller must block sigmask_newstate signals.
 */
void
newstate(action_t *camreqlog)
{
	action_t req, *camreq, camreqbuf[CONFCAMMAX];
	uid_t uid;
	pid_t pid;
	int i, j, didcap;

	openpid();
	pid = readpid(pid_req, camreqbuf, &uid);
	if (camreqlog == NULL) {
		if (pid == 0) {			/* no request was pending */
			closepid();
			return;
		}
		camreq = camreqbuf;
	} else {
		/*
		 * It's possible that a client has a pending request.
		 * If that happened (and not going down on a SIGTERM),
		 * ignore log-driven request to service the client.
		 * The log-driven event will call back soon enough.
		 */
		if (pid == 0 || camreqlog == (action_t *)1) {
			if (camreqlog == (action_t *)1) {
				for (i = 0; i < TotCam; i++)
					camreqbuf[i] = act_down;
				camreq = camreqbuf;
			} else {
				camreq = camreqlog;
			}
			pid = getpid();
			uid = MyUid;
		} else				/* override log-driven */
			camreq = camreqbuf;
	}

	/*
	 * 'camreq' holds the list of target states.
	 *
	 * If not bringing everything down, ensure image directory
	 * is (still) mounted.
	 */
	req = act_down;
	for (i = 0; i < TotCam; i++)
		if (camreq[i] != act_none && camreq[i] != act_down) {
			req = camreq[i];
			break;
		}
	if (req != act_down && mntexec(mnt_on) == 0)	/* mount failed... */
		for (i = 0; i < TotCam; i++)
			camreq[i] = act_down;		/* shut down cameras */

	/*
	 * Walk 'ovride' (override list) for cameras transitioning into
	 * capture mode.  Increment each overriden camera's 'ovcnt' and,
	 * if already in capture mode, kill their capture process.
	 */
	for (i = 0; i < TotCam; i++) {
		if (Cf->camera[i].ovdone)	/* override already in effect */
			continue;
		if (camreq[i] == act_cap || camreq[i] == act_autocap) {
			struct confcam_s *cfc = Cf->camera[i].ovride;
			while (cfc != NULL) {
				j = cfc - Cf->camera;
				if (cfc->ovcnt++ == 0 && CapPid[j] != 0) {
					syslog(LOG_INFO, "%s: overridden by %s",
						cfc->name, Cf->camera[i].name);
					(void) capexec(cap_stop, j, 0);
					(void) kill(CapPid[j], SIGTERM);
					CapPid[j] = 0;
				}
				cfc = cfc->ovride;
			}
			Cf->camera[i].ovdone = 1;
		}
	}
	/*
	 * Walk 'ovride' (override list) for cameras transitioning out of
	 * capture mode.  Decrement each overriden camera's 'ovcnt' and,
	 * if it becomes 0 in capture mode, restart their capture process.
	 */
	for (i = 0; i < TotCam; i++) {
		if (! Cf->camera[i].ovdone)	/* no overrides in effect */
			continue;
		if (camreq[i] == act_up || camreq[i] == act_down) {
			struct confcam_s *cfc = Cf->camera[i].ovride;
			while (cfc != NULL) {
				j = cfc - Cf->camera;
				if (--cfc->ovcnt == 0 &&
						(CurState[j] == act_cap ||
						 CurState[j] == act_autocap)) {
					CapPid[j] = capexec(cap_start, j, 0);
					syslog(LOG_INFO, "%s: restarted by %s",
						cfc->name, Cf->camera[i].name);
				}
				cfc = cfc->ovride;
			}
			Cf->camera[i].ovdone = 0;
		}
	}

	/* carry out the requested action(s) */
	for (i = 0, didcap = 0, req = act_down; i < TotCam; i++) {
		switch(camreq[i]) {
		    case act_up:		/* kill capture */
		    case act_down:
			if (CapPid[i] != 0) {
				(void) capexec(cap_stop, i, 0);
				(void) kill(CapPid[i], SIGTERM);
				CapPid[i] = 0;
			}
			/*
			 * When auto-transitioning out of capture,
			 * we need to get the default image back
			 * up there.  Also, any related connection
			 * information must be cleared.
			 *
			 * An "odd" situation occurs when the admin
			 * requests 'act_autocap' -> 'act_up'; Why
			 * do this?  If someone is pulling frames,
			 * we'll just auto-transition back.  Assume
			 * they want to start fresh and clear out
			 * any related connection information.
			 */
			writedefimg(&Cf->camera[i]);
			climgr(cli_clrcam, (camidx_t)i, NULL, NULL, NULL);
			break;

		    case act_cap:
			/* act_autocap -> act_cap: reset sampling */
			if (CurState[i] == act_autocap) {
				capexec(cap_nsamp, i, Cf->capival);
				climgr(cli_rsamp, i, NULL, NULL, NULL);
			}
			/*FALLTHRU*/

		    case act_autocap:		/* start capture */
			if (CapPid[i] == 0 && Cf->camera[i].ovcnt == 0) {
				if (didcap)	/* stagger starts */
					usleep(200000);
				CapPid[i] = capexec(cap_start, i, 0);
				if (CapPid[i] == 0)	/* failed */
					camreq[i] = act_up;
				else
					didcap = 1;
			}
			break;

		    default:
			break;
		}
		if (camreq[i] != act_none) {	/* update CurState[], logging */
			syslog(LOG_NOTICE, "%s: state [%s] -> [%s] (uid:%u)",
				Cf->camera[i].name, actmsg[CurState[i]],
				actmsg[camreq[i]], uid);
			CurState[i] = camreq[i];
		}
		if (CurState[i] != act_down)	/* prevent exit if !act_down */
			req = CurState[i];
	}

	writepid(pid_cur, NULL, pid);

	if (req == act_down) {
		(void) mntexec(mnt_off);	/* ensure umounted */
		syslog(LOG_NOTICE, "server exiting...");
		closepid();
		exit(0);
	}

	closepid();
}

/*
 * acthdlr(sig) - signal handler for state-change requests.
 *
 * Once the server is running, clients send us SIGUSR1 ('sig') to
 * initiate state change transitions.  On SIGTERM, clean up and
 * exit (gracefully).
 */
void
acthdlr(int sig)
{
	switch(sig) {
	    case SIGUSR1:
		newstate(NULL);
		break;
	    case SIGTERM:
		newstate((action_t *)1);
		break;
	    default:
		syslog(LOG_NOTICE, "acthdlr: unexpected signal: %d", sig);
		break;
	}
}

/*
 * reaper(sig) - SIGCHLD handler (child reaper).
 *
 * Reap exit status from children; actual status doesnt matter,
 * just dont want to have zombies hanging around.  Also check
 * that we didnt unexpectedly lose a capture process; if so,
 * clear it from CapPid[] and set camera state back to 'act_up'.
 */
void
reaper(int sig)
{
	const char *anyex = "child %d exited w/0x%x";
	const char *capex = "%s: state [%s] -> [%s] (capture %d exited w/0x%x)";
	int status, i;
	pid_t pid;

	while ((pid = wait3(&status, WNOHANG, NULL)) > 0) {
		for (i = 0; i < CONFCAMMAX; i++)
			if (CapPid[i] == pid) {
				syslog(LOG_WARNING, capex, Cf->camera[i].name,
					actmsg[CurState[i]], actmsg[act_up],
					pid, status);
				capexec(cap_stop, i, 0);	/* close pipe */
				CapPid[i] = 0;
				CurState[i] = act_up;
				openpid();
				writepid(pid_curonly, NULL, (pid_t)0);
				closepid();
				break;
			}
		if (i == CONFCAMMAX)
			syslog(LOG_DEBUG, anyex, pid, status);
	}
}

/*
 * caphdlr(sig) - signal handler for capture manager process.
 */
void
caphdlr(int sig)
{
	pid_t pid;
	int status;

	switch(sig) {
	    case SIGCHLD:
		/*
		 * If the child capture process exited, send the entire
		 * process group a SIGTERM.  Also squirrel away the exit
		 * status in 'capman_stat' and clear 'capman_pid' to let
		 * the capture manager know the child exited.
		 *
		 * If 'capman_pid' equals -2, the capture manager
		 * already knows, so just exit now.
		 */
		while ((pid = wait3(&status, WNOHANG, NULL)) > 0) {
			if (capman_pid == -2)	/* capture manager knows */
				_exit(0);
			if (pid != capman_pid)
				continue;
			killpg(capman_pid, SIGTERM);
			capman_stat = status;
			capman_pid = 0;
		}
		break;
	    case SIGHUP:
	    case SIGINT:
	    case SIGQUIT:
	    case SIGTERM:
	    case SIGPIPE:
		/*
		 * If we (the capture manager) receive a SIGTERM, cammgr
		 * is asking us to shut down the capture process and exit.
		 *
		 * Receiving any of the other signals listed above imply
		 * a shutdown of the capture process as well.
		 */
		if (capman_pid > 0)
			killpg(capman_pid, SIGTERM);
		_exit(0);
	    default:
		syslog(LOG_NOTICE, "caphdlr: unexpected signal: %d", sig);
		break;
	}
}


/*
 * strmatch(opt, arg) - partial string matcher.
 *
 * Return true if 'arg' is short (or identical) to 'opt' string.
 */
int
strmatch(const char *opt, const char *arg)
{
	if (*arg != '\0') {
		while (*opt != '\0' && *opt == *arg)	/* while same... */
			opt++, arg++;
		if (*arg == '\0')			/* have a match */
			return 1;
	}
	return 0;
}

/*
 * usage(progname, errmsg) - print error/usage info to STDERR and return '1'.
 */
int
usage(const char *progname, const char *errmsg)
{
	fprintf(stderr, "%s: parse error: %s\n", progname, errmsg);
	fprintf(stderr, "usage: %s [-fvw] [ up | capture | down ] [ <camera1> ... <cameraN> ]\n", progname);
	return 1;
}

/*
 * version() - print version to STDOUT and return '0'.
 */
int
version(void)
{
	printf(CopyrightVersion, Version,
#ifdef MNTDIR
					" +MNTDIR"
#else
					""
#endif
					,
#ifdef CAPFUN
					" +CAPFUN"
#else
					""
#endif
					);
	return 0;
}

/*
 * MKCONDIR(wmsg, dir)
 *
 * Try twice to create the connection directory 'dir' (Cf->curcondir).
 * If the first attempt fails because it already exists, remove
 * and then try a second time.  If that fails, or if mkdir()
 * reports any other problem, log a warning (this isnt fatal).
 */
#define	MKCONDIR(wmsg, dir) if (dir != NULL) {				\
	int _retrycnt;							\
	for (_retrycnt = 0; _retrycnt < 2; _retrycnt++) {		\
		if (mkdir(dir, 0755) < 0) {				\
			if (errno == EEXIST) {				\
				removedir(dir);				\
				continue;				\
			}						\
			syslog(LOG_WARNING, wmsg, dir);			\
		}							\
		break;							\
	}								\
}

/*
 * mntexec(op) - mount/unmount image directory (if MNTDIR defined),
 *               ensure that directory exists (if MNTDIR not defined).
 *
 * Return non-zero if 'Cf->curimgdir' is available, 0 on failure.
 * If a fork() required, child pid is returned, else return is '1'.
 *
 * If 'Cf->curcondir' defined, ensure that the connection directory
 * exists and is empty.  Also write out default images after a
 * successful mount/unmount.
 */
pid_t
mntexec(optmnt_t op)
{
	const char *noconnmsg = "no connection updates: %s: %m";
#ifndef	MNTDIR
	struct stat sb;
	static int mounted = 0;

	/*
	 * If not doing mount/unmount, all that's necessary for success
	 * is that the target directory exists.
	 */
	if (stat(Cf->curimgdir, &sb) < 0)
		syslog(LOG_ERR, "stat(%s): %m", Cf->curimgdir);
	else if ((sb.st_mode & S_IFMT) != S_IFDIR)
		syslog(LOG_ERR, "stat(%s): not a directory", Cf->curimgdir);
	else {
		/* create or remove connection directory depending on 'op' */
		switch(op) {
		    case mnt_on:
			if (mounted == 0) {
				writedefimg(NULL);
				MKCONDIR(noconnmsg, Cf->curcondir);
				mounted = 1;
			}
			break;
		    case mnt_off:
			if (mounted == 1) {
				writedefimg(NULL);
				if (Cf->curcondir != NULL)
					removedir(Cf->curcondir);
				mounted = 0;
			}
			break;
		    case mnt_offloop:		/* impossible */
			break;
		}
		return 1;	/* success */
	}
	return 0;		/* failure */
#else
#define	TENTHSECS	150
	static char *const mntargs[] = { CAMMNT, NULL };
	static char *const umntargs[] = { CAMMNT, "-u", NULL };
	char *const *argv = (op == mnt_on)? mntargs: umntargs;
	sigset_t osigmask;
	const char *mntpnt;
	int i, cmp, pidcnt = 0;
	pid_t pid, pidkids[TENTHSECS/10 + 1];

	if ((mntpnt = getmntpnt(Cf->curimgdir)) == NULL)
		return 0;

	/* if already mounted or unmounted, return successful */
	cmp = strcmp(Cf->curimgdir, mntpnt);
	if ((op == mnt_on && cmp == 0) ||
			((op == mnt_off || op == mnt_offloop) && cmp != 0))
		return 1;

	bzero(pidkids, sizeof pidkids);
	sigprocmask(SIG_BLOCK, &sigmask_newstate, &osigmask);
	switch(pid = fork()) {
	    case -1:		/* error */
		syslog(LOG_ERR, "fork(%s): %m", argv[0]);
		pid = 0;
		break;

	    case 0:		/* child */
		closepid();			/* parent still holds lock */
		capexec(cap_closeall, 0, 0);	/* close any pipes */
		CLOSE_LOGFD();			/* close event-driven log */
		if ((i = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
			(void) dup2(i, STDIN_FILENO);
			(void) dup2(i, STDOUT_FILENO);
			(void) dup2(i, STDERR_FILENO);
		}
#ifdef	HAS_SETPROCTITLE
		setproctitle(mntmsg[op]);
#endif
		SIGRESET();
		syslog(LOG_DEBUG, "execve(%s)", argv[0]);
		execve(argv[0], argv, NULL);
		syslog(LOG_ERR, "execve(%s): %m", argv[0]);
		closelog();
		_exit(1);

	    default:		/* parent */
		if (op == mnt_offloop)		/* caller already waiting */
			return pid;

		/* wait (TENTHSECS/10) seconds for operation to complete... */
		for (i = 0; i < TENTHSECS; i++) {
			if ((mntpnt = getmntpnt(Cf->curimgdir)) != NULL) {
				cmp = strcmp(Cf->curimgdir, mntpnt);
				if ((op == mnt_on && cmp == 0) ||
				    (op == mnt_off && cmp != 0))
					break;
			}
			usleep(100000);

			/* if unmount: try, try again... */
			if (op == mnt_off && ((i % 10) == 9))
				pidkids[pidcnt++] = mntexec(mnt_offloop);
		}
		if (i == 150) {
			syslog(LOG_ERR, "%s timeout (%s): giving up.",
				mntmsg[op], Cf->curimgdir);

			/* try to clean up by killing children */
			(void) kill(pid, SIGTERM);
			for (i = 0; i < pidcnt; i++)
				if (pidkids[i] > 1)
					(void) kill(pidkids[i], SIGTERM);
			sleep(3);
			(void) kill(pid, SIGKILL);
			for (i = 0; i < pidcnt; i++)
				if (pidkids[i] > 1)
					(void) kill(pidkids[i], SIGKILL);
			pid = 0;
			break;
		}

		/*
		 * Immediately write default images to target directory.
		 * If doing a mount, the initial images are created.
		 * If doing a umount, the default on-disk images are
		 * rewritten so the web server will refresh on change.
		 */
		writedefimg(NULL);

		/* create or remove connection directory depending on 'op' */
		if (Cf->curcondir != NULL) {
			if (op == mnt_off)
				removedir(Cf->curcondir);
			else
				MKCONDIR(noconnmsg, Cf->curcondir);
		}
		break;
	}

	sigprocmask(SIG_SETMASK, &osigmask, NULL);
	return pid;
#undef	TENTHSECS
#endif
}

#define	CAPEXEC_POLLSEC	1

/*
 * capexec(cmd, camidx, msec) - manage capture sub-process.
 *
 * Where 'cmd' is one of:
 *	cap_start    - Launch the set of commands that capture data
 *	               to the image files for camera 'camidx'.
 *	cap_nsamp    - Send new sampling interval ('msec' milliseconds)
 *	               to capture process for camera 'camidx'.
 *	cap_stop     - Terminate capture process for camera 'camidx'.
 *	cap_closeall - Close all pipe descriptors (used after fork()).
 *
 * Return child pid on success, 0 on failure.
 *
 * If compiled with WAIT_CAPEXIT, capexec(cap_stop) waits a full sample
 * interval for child to read from pipe and exit cleanly.  Remember,
 * we are holding the PID lock here too.  In practice, WAIT_CAPEXIT
 * is only necessary on systems that do not support forcible unmounts,
 * where the unmount fails because the capexec process holds the file
 * system open.  Of course, the occasional http requests will do that
 * too, so it's really not a good idea to do the mount/unmount thing
 * without the forced unmount feature, is it?
 *
 * The value CAPEXEC_POLLSEC (defined above) is used only for continously
 * running capture processes (i.e., those that start with a bang (!)).
 * It's specifies the amount of time between polls to see that the capture
 * process is still healthy.  There should be no reason to change it.
 */
pid_t
capexec(capcmd_t cmd, camidx_t camidx, unsigned int msec)
{
	const struct confcam_s *cfc = &Cf->camera[camidx];
	struct sigaction act;
	sigset_t osigmask, nsigmask;
	static int firstime = 1, cappipes[CONFCAMMAX][2];
#ifdef	WAIT_CAPEXIT
	static int savcapival[CONFCAMMAX];
#endif
	const char *capcmd;
	int i, curcapival, capcontrun, exstat, *pipefd = cappipes[camidx];
	pid_t pid;
#define	PCLOSE(fd)	do { (void) close(fd); fd = -1; } while(0)

	/*
	 * It is not possible to call capexec() with a 'camidx' that
	 * points to an uninitialized pipe in cappipes[].  However,
	 * since the "uninitialized" value is "0", let's ensure that
	 * everything is "-1" so we dont screw with a valid descriptor.
	 */
	if (firstime) {
		for (i = 0; i < CONFCAMMAX; i++)
			cappipes[i][0] = cappipes[i][1] = -1;
		firstime = 0;
	}

	/* set true if capture operation runs continuously */
	capcontrun = (cfc->capcmd != NULL && cfc->capcmd[0] == '!');

	if (cmd != cap_start) {
		switch (cmd) {
		    case cap_nsamp:
			if (pipefd[1] >= 0) {
				(void) write(pipefd[1], &msec, sizeof msec);
#ifdef	WAIT_CAPEXIT
				savcapival[camidx] = msec;
#endif
			}
			break;
		    case cap_stop:
			if (pipefd[1] >= 0) {
#ifdef	WAIT_CAPEXIT
				unsigned int secs = savcapival[camidx] / 1000;
				unsigned int usecs =
					(savcapival[camidx] % 1000) * 1000;
#endif
				PCLOSE(pipefd[1]);
#ifdef	WAIT_CAPEXIT
				if (capcontrun && secs < (CAPEXEC_POLLSEC * 3))
					secs = CAPEXEC_POLLSEC * 3;
				if (secs > 0)
					sleep(secs);
				if (usecs > 0)
					usleep(usecs);
#else
				/*
				 * For continuous-run capture processes,
				 * give time to notify childen to exit.
				 */
				if (capcontrun)
					sleep(CAPEXEC_POLLSEC * 3);
#endif
			}
			break;
		    case cap_closeall:
			for (i = 0; i < CONFCAMMAX; i++) {
				if (cappipes[i][0] != -1)
					PCLOSE(cappipes[i][0]);
				if (cappipes[i][1] != -1)
					PCLOSE(cappipes[i][1]);
			}
			break;
		    default:			/* impossible */
			break;
		}
		return 1;
	}

	/* create a non-blocking pipe between us and the capture process */
	if (pipe(pipefd) < 0) {
		pipefd[0] = pipefd[1] = -1;
	} else {
		int fl0, fl1;

		/* if we cant get non-blocking, give up on the pipe */
		if ((fl0 = fcntl(pipefd[0], F_GETFL, 0)) < 0 ||
		    (fl1 = fcntl(pipefd[1], F_GETFL, 0)) < 0 ||
		    fcntl(pipefd[0], F_SETFL, fl0 | O_NONBLOCK) < 0 ||
		    fcntl(pipefd[1], F_SETFL, fl1 | O_NONBLOCK) < 0) {
			syslog(LOG_WARNING, "%s: fcntl(pipe, O_NONBLOCK): %m",
				cfc->name);
			PCLOSE(pipefd[0]);
			PCLOSE(pipefd[1]);
		} else {
			/* set close-on-exec, at least for system(3) */
			(void) fcntl(pipefd[0], F_SETFD, 1);
			(void) fcntl(pipefd[1], F_SETFD, 1);
		}
	}

	sigprocmask(SIG_BLOCK, &sigmask_newstate, &osigmask);

	/* reset sampling interval before launching capture */
#ifdef	WAIT_CAPEXIT
	savcapival[camidx] =		/* ... curcapival (from next line) */
#endif
	curcapival = Cf->capival;	/* look up... look up one line! */
	climgr(cli_rsamp, camidx, NULL, NULL, NULL);

	switch(pid = fork()) {
	    case -1:
		syslog(LOG_ERR, "%s: fork(capture): %m", cfc->name);
		pid = 0;
		PCLOSE(pipefd[0]);
		PCLOSE(pipefd[1]);
		break;

	    case 0:		/* child */
		/*
		 * This is the client capture manager!
		 *
		 * This block of code is responsible for running the
		 * image capture operation for a particular camera.
		 *
		 * Close all unnecessary file descriptors, chdir to the
		 * camera image directory and set up a whole new set of
		 * signal handlers.
		 */
		closepid();			/* parent still holds lock */
		CLOSE_LOGFD();			/* close event-driven log */
#ifdef	HAS_SETPROCTITLE
		setproctitle("capture: %s", cfc->name);
#endif
		/* close all pipes *except* our read end! */
		for (i = 0; i < CONFCAMMAX; i++) {
			if (i != camidx && cappipes[i][0] != -1)
				PCLOSE(cappipes[i][0]);
			if (cappipes[i][1] != -1)
				PCLOSE(cappipes[i][1]);
		}
		if (chdir(Cf->curimgdir) < 0) {
			syslog(LOG_ERR, "%s: chdir(%s): %m", cfc->name,
				Cf->curimgdir);
			_exit(1);
		}

		pid = 0;
		if (capcontrun) {
			capcmd = cfc->capcmd + 1;
			capman_pid = pid;
		} else {
			capcmd = cfc->capcmd;
			capman_pid = -1;
		}

		SIGRESET();
		sigemptyset(&nsigmask);
		sigaddset(&nsigmask, SIGHUP);
		sigaddset(&nsigmask, SIGINT);
		sigaddset(&nsigmask, SIGQUIT);
		sigaddset(&nsigmask, SIGTERM);
		sigaddset(&nsigmask, SIGPIPE);
		sigaddset(&nsigmask, SIGCHLD);

		act.sa_flags = 0;
		act.sa_mask = nsigmask;
		act.sa_handler = caphdlr;
		sigaction(SIGHUP, &act, (struct sigaction *)NULL);
		sigaction(SIGINT, &act, (struct sigaction *)NULL);
		sigaction(SIGQUIT, &act, (struct sigaction *)NULL);
		sigaction(SIGTERM, &act, (struct sigaction *)NULL);
		sigaction(SIGPIPE, &act, (struct sigaction *)NULL);
		sigaction(SIGCHLD, &act, (struct sigaction *)NULL);

		/* does administrator request delay before starting capture? */
		if (Cf->capdelay)
			sleep(Cf->capdelay);

		/*
		 * Capture commands beginning with a bang (!) are used to
		 * indicate continuously running capture process.  Here,
		 * we fork(), where the parent reads the pipe and ensures
		 * that the child capture process does not die.  We have
		 * no control over the sampling interval for these cameras.
		 *
		 * The other method for capturing is to launch an program
		 * each time to grab an image.  Here, we DO have control
		 * over the sampling interval.
		 *
		 * The following loop handles both capture methods.
		 */
		for (;;) {
			struct timeval tv_s, tv_f, tv_e;
			char * const *l, imgfil[CONFIMGLEN+1];	/* +1 for '_' */
			ssize_t nr;

			gettimeofday(&tv_s, NULL);

			/*
			 * When the other end of the pipe is closed,
			 * shut down the capture process and exit
			 * with the specified 'exstat' status.
			 */
			if ((nr = read(pipefd[0], &msec, sizeof msec)) < 0) {
				if (errno != EAGAIN) {
					syslog(LOG_ERR, "%s: read(pipe): %m",
							cfc->name);
					exstat = 2;
					break;
				}
			} else if (nr == 0) {
				exstat = 0;
				break;
			} else if (!capcontrun && nr == sizeof msec) {
				curcapival = msec;
				syslog(LOG_INFO, "%s: capival %ums.",
					cfc->name, curcapival);
			}

			/*
			 * If a continuously running capture exits,
			 * log a message to alert the administrator.
			 */
			if (capcontrun && pid != capman_pid) {
				if (WIFEXITED(capman_stat)) {
					syslog(LOG_NOTICE,
						"%s: capture exit (%u)",
						cfc->name, pid);
				} else if (WIFSIGNALED(capman_stat)) {
					syslog(LOG_NOTICE,
						"%s: capture died (%u)",
						cfc->name, pid);
				}
				pid = 0;	/* set to relaunch capture */
			}

			/* run internal capture command (if configured) */
			if (cfc->capfun != NULL)
				(*cfc->capfun)(cfc);

			/* run external capture command (if configured) */
			if (cfc->capcmd != NULL && pid == 0) {
 				/*
				 * Note that the child is made a session
				 * leader so that the parent can killpg()
				 * all the spawned processes if necessary.
				 */
				sigprocmask(SIG_BLOCK, &nsigmask, &osigmask);
				switch(capman_pid = fork()) {
				    case -1:
					syslog(LOG_ERR, "fork(%s): %m", capcmd);
					sleep(30);
					capman_pid = 0;
					break;
				    case 0:		/* child */
					SIGRESET();
					(void) setsid();
					PCLOSE(pipefd[0]);
					closelog();
					if ((i = open(_PATH_DEVNULL,
							O_RDWR, 0)) != -1) {
						dup2(i, STDIN_FILENO);
						dup2(i, STDOUT_FILENO);
						dup2(i, STDERR_FILENO);
					}
					execl(_PATH_BSHELL, "sh", "-c",
							capcmd, (char *)NULL);
					_exit(127);
				    default:		/* parent */
					break;
				}
				if (capcontrun)
					pid = capman_pid;
				else
					while (capman_pid != 0)
						sigsuspend(&osigmask);
				sigprocmask(SIG_SETMASK, &osigmask, NULL);
			}

			/*
			 * For continuously running captures, we need a
			 * timeout before looping again to check the status.
			 * It would be nice to sigsuspend(), but we have to
			 * read from that pipe.
			 *
			 * Otherwise, determine how long to sleep before
			 * the next image grab.
			 */
			if (capcontrun) {
				sleep(CAPEXEC_POLLSEC);
			} else {
				for (l = cfc->imgfil; l < cfc->imgfil_e; l++) {
					*imgfil = '_';
					strcpy(imgfil + 1, *l);
					(void) rename(imgfil, *l);
				}

				/* how long to sleep for next iteration? */
				gettimeofday(&tv_f, NULL);
				timersub(&tv_f, &tv_s, &tv_e);
				tv_s.tv_sec = curcapival / 1000;
				tv_s.tv_usec = (curcapival % 1000) * 1000;
				if (timercmp(&tv_s, &tv_e, > )) {
					timersub(&tv_s, &tv_e, &tv_f);
					if (tv_f.tv_sec > 0)
						sleep(tv_f.tv_sec);
					if (tv_f.tv_usec > 0)
						usleep(tv_f.tv_usec);
				}
			}
		}
		syslog(LOG_INFO, "%s: terminating capture (pgrp:%u)",
			cfc->name, capman_pid);
		if (capman_pid > 0) {
			pid = capman_pid;
			capman_pid = -2;
			killpg(pid, SIGTERM);
			sleep(3);
		}
		_exit(exstat);
		/*NOTREACHED*/

	    default:		/* parent */
		PCLOSE(pipefd[0]);		/* close our read-side */
		break;
	}

	sigprocmask(SIG_SETMASK, &osigmask, NULL);
	return pid;
#undef	PCLOSE
}

/*
 * writedefimg(cam) - write default images.
 *
 * Write out the default images for camera 'cam'.  If 'cam' is NULL,
 * write out default images for every camera.  Also remove temp image
 * files (i.e. those with a leading underscore) for targetted camera(s).
 */
void
writedefimg(struct confcam_s *cam)
{
	static char fil[PATH_MAX], *filnam;
	static int remcnt = -1;
	struct confcam_s *cfc;
	char **clst, **dlst;
	size_t *dsiz;
	FILE *fp;

	/*
	 * Initialize 'fil' to path to Cf->curimgdir.
	 * Position 'filnam' to byte after trailing slash.
	 * Ultimately set 'remcnt' to space remaining in 'fil'.
	 */
	if (remcnt == -1) {
		snprintf(fil, PATH_MAX, "%s/", Cf->curimgdir);
		remcnt = strlen(fil);
		filnam = fil + remcnt;
		remcnt = PATH_MAX-1 - remcnt;
	}

	/* walk the cameras; if they have a default image, write it out */
	for (cfc = (cam == NULL)? Cf->camera: cam; cfc < Cf->camera_e; cfc++) {
		for (clst = cfc->imgfil, dlst = cfc->defimg,
				dsiz = cfc->defsiz; clst < cfc->imgfil_e;
				clst++, dlst++, dsiz++) {
			if (*dlst != NULL) {
				strncpy(filnam, *clst, remcnt);
				if ((fp = fopen(fil, "w")) != NULL) {
					fwrite(*dlst, sizeof(char), *dsiz, fp);
					(void) fclose(fp);
				}
			}
			/* remove temp image file */
			snprintf(filnam, remcnt, "_%s", *clst);
			(void) unlink(fil);
		}
		if (cam != NULL)	/* single pass (targetted camera) */
			break;
	}
}

/*
 * cliexec(camera, user, addr, arg) - launch per-camera/client command
 *
 * If 'Cf->cliapp' is non-NULL, invoke it with the arguments:
 *
 *    <user> <camera> <addr> ( connect | drop )
 *
 * Return child pid if exec'd, 0 if nothing to do (or error).
 */
pid_t
cliexec(const char *camera, const char *user, const char *addr, const char *arg)
{
	sigset_t osigmask;
	const char *argv[6];
	struct stat sb;
	pid_t pid;
	int fd;

	if (Cf->cliapp == NULL)		/* no client command */
		return 0;

	if (stat(Cf->cliapp, &sb) < 0) {
		syslog(LOG_WARNING, "%s: %m", Cf->cliapp);
		return 0;
	}

	if ((sb.st_mode & S_IFMT) != S_IFREG ||
	    (sb.st_mode & (S_IRUSR|S_IXUSR)) != (S_IRUSR|S_IXUSR)) {
		syslog(LOG_WARNING, "%s: bad protection mode: %x",
			Cf->cliapp, sb.st_mode);
		return 0;
	}

	argv[0] = Cf->cliapp;
	argv[1] = user;
	argv[2] = camera;
	argv[3] = addr;
	argv[4] = arg;
	argv[5] = NULL;

	sigprocmask(SIG_BLOCK, &sigmask_newstate, &osigmask);
	switch(pid = fork()) {
	    case -1:
		syslog(LOG_ERR, "fork(%s): %m", argv[0]);
		pid = 0;
		break;

	    case 0:		/* child */
		closepid();			/* parent still holds lock */
		capexec(cap_closeall, 0, 0);	/* close any pipes */
		CLOSE_LOGFD();			/* close event-driven log */
		if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
			(void) dup2(fd, STDIN_FILENO);
			(void) dup2(fd, STDOUT_FILENO);
			(void) dup2(fd, STDERR_FILENO);
		}
#ifdef	HAS_SETPROCTITLE
		setproctitle("cliexec: %s: %s", camera, user);
#endif
		SIGRESET();
		syslog(LOG_DEBUG, "execve(%s,%s,%s,%s,%s)",
			argv[0], user, camera, addr, arg);
		execve(Cf->cliapp, (char * const *)argv, NULL);
		syslog(LOG_ERR, "execve(%s): %m", argv[0]);
		closelog();
		_exit(1);

	    default:		/* parent */
		break;
	}

	sigprocmask(SIG_SETMASK, &osigmask, NULL);
	return pid;
}

/*
 * climgr(cmd, tci, user, addr, changes) - manage client connection dir.
 *
 * Where 'cmd' is one of:
 *	cli_add    - add a client (or update its internal timestamp)
 *	             'user' is the username, 'addr' is IP addr string.
 *	             also do a 'cli_prune' while we are here.
 *	cli_prune  - prune existing clients.
 *	cli_clrcam - clear all existing clients for camera 'camidx'.
 *	cli_clrall - clear all existing clients.
 *	cli_rsamp  - reset sampling interval to Cf->capival.
 *                   return 0 regardless of actual client count.
 *
 * When command is 'cli_add', 'cli_clrcam', or 'cli_rsamp', the argument
 * 'tci' identifies the target camera.  Otherwise, 'tci' is ignored.
 *
 * Return number of clients still connected and, if 'changes' is non-NULL,
 * return the set of cameras that changes states.
 *
 * N.B. Yes, I'm currently only using a singularly linked list.  If you
 * have bunches of users and cameras, consider making some changes here!
 */
int
climgr(clicmd_t cmd, camidx_t tci, const char *user, const char *addr,
		action_t *changes)
{
#define	MAXCLINLEN	64
#define	MAXADDRLEN	64
#define	MAXTIMESAM	8
	static char fil[PATH_MAX], *filnam;
	static int remcnt = -1;
	static struct camcli_s {
		struct camcli_s *next;
		time_t prevtime;
		time_t currtime;
		time_t conntime;
		int camidx;
		char name[MAXCLINLEN];
		char addr[MAXADDRLEN];
	} *cchead = NULL, *ccfree = NULL;
	static struct sample_s {
		time_t samples[MAXTIMESAM];
		int sampidx;
		int curcapival, clicnt;
		time_t tdiff;
	} sample[CONFCAMMAX];
	struct camcli_s *cc, *cc_p;
	struct sample_s *sp, *sp_e;
	sigset_t osigmask;
	time_t tvec;
	camidx_t ci;
	char *camnam;
	int totcli = 0;

	/* prevent asynchronous state transitions while we walk the list */
	sigprocmask(SIG_BLOCK, &sigmask_newstate, &osigmask);

	if (remcnt == -1) {		/* first time here */
		/*
		 * Initialize 'fil' to path to Cf->curcondir.
		 * Position 'filnam' to byte after trailing slash.
		 * Ultimately set 'remcnt' to space remaining in 'fil'.
		 */
		if (Cf->curcondir != NULL) {
			snprintf(fil, PATH_MAX, "%s/", Cf->curcondir);
			remcnt = strlen(fil);
			filnam = fil + remcnt;
			remcnt = PATH_MAX-1 - remcnt;
		}
		/* initialize all camera capture intervals */
		for (sp = sample, sp_e = &sample[CONFCAMMAX]; sp < sp_e; sp++)
			sp->curcapival = Cf->capival;
	}

	if (cmd == cli_rsamp) {	/* reset this camera's capture interval */
		sample[tci].curcapival = Cf->capival;
		goto rsampout;
	}

	time(&tvec);

	/* adding (or updating) a client */
	if (cmd == cli_add) {
		/* is this an existing client? */
		for (cc = cchead; cc != NULL; cc = cc->next)
			if (tci == cc->camidx &&
			    strncmp(user, cc->name, MAXCLINLEN-1) == 0 &&
			    strncmp(addr, cc->addr, MAXADDRLEN-1) == 0) {
				cc->prevtime = cc->currtime;
				cc->currtime = tvec;
				goto prune;
			}

		/* not an existing client, so add them */
		if (ccfree == NULL) {
			if ((cc = malloc(sizeof(struct camcli_s))) == NULL) {
				syslog(LOG_WARNING,
					"climgr: malloc(%u): out of memory",
					(unsigned) sizeof(struct camcli_s));
				goto prune;
			}
		} else
			cc = ccfree, ccfree = ccfree->next;

		cc->prevtime = 0;
		cc->conntime = cc->currtime = tvec;
		cc->camidx = tci;
		strncpy(cc->name, user, MAXCLINLEN);
		cc->name[MAXCLINLEN-1] = '\0';
		strncpy(cc->addr, addr, MAXADDRLEN);
		cc->addr[MAXADDRLEN-1] = '\0';
		camnam = Cf->camera[cc->camidx].name;

		if (Cf->curcondir != NULL) {
			int fd;
			snprintf(filnam, remcnt, "%s_%s_%s", camnam,
				cc->name, cc->addr);
			if ((fd = open(fil, O_WRONLY|O_CREAT, 0644)) >= 0)
				(void) close(fd);
		}

		cc->next = (cchead == NULL)? NULL: cchead;
		cchead = cc;

		syslog(LOG_NOTICE, "conn %s:%s:%s", camnam, cc->name, cc->addr);
		cliexec(camnam, cc->name, cc->addr, "connect");
	}

prune:
	for (sp = sample, sp_e = &sample[TotCam]; sp < sp_e; sp++) {
		sp->tdiff = Cf->usertout * 8;		/* something big */
		sp->clicnt = 0;
	}
restart:
	for (cc_p = NULL, cc = cchead; cc != NULL; cc_p = cc, cc = cc->next) {
		ci = cc->camidx;
		if ((cmd == cli_clrall) ||
				(cmd == cli_clrcam && tci == ci) ||
				(tvec > cc->currtime + Cf->usertout)) {

			/* this client is no longer watching this camera */
			camnam = Cf->camera[ci].name;

			if (cc_p == NULL)
				cchead = cc->next;
			else
				cc_p->next = cc->next;

			if (Cf->curcondir != NULL) {
				snprintf(filnam, remcnt, "%s_%s_%s",
						camnam, cc->name, cc->addr);
				(void) unlink(fil);
			}

			cc->next = (ccfree == NULL)? NULL: ccfree;
			ccfree = cc;

			syslog(LOG_NOTICE, "drop %s:%s:%s (%lu minutes)",
				camnam, cc->name, cc->addr, (unsigned long)
				((tvec - cc->conntime) / 60) + 1);

			cliexec(camnam, cc->name, cc->addr, "drop");

			/* reinitialize 'cc' and continue walking the list */
			if (cc_p == NULL)	/* removed first element... */
				goto restart;		/* just start again */
			cc = cc_p;
		} else {
			sp = &sample[ci];
			if (cc->prevtime != 0) {
				time_t td = cc->currtime - cc->prevtime;
				if (td < sp->tdiff)
					sp->tdiff = td;
			}
			sp->clicnt++;
			totcli++;
		}
	}

	/*
	 * Say a client is pulling an image every 10 seconds, why waste
	 * time taking a picture every second?  I can think of 3 reasons:
	 *   1) You want what ever picture they get to be most current,
	 *   2) Your camera has a light that blinks on each image grab;
	 *      when you look at the camera, you can quickly determine
	 *      if someone is watching because of the blinking light.
	 *   3) Your camera/driver auto-adjusts for light and, if unused
	 *      for a short period of time, it resets and loses state.
	 *
	 * If none of these conditions apply and you use HTTP requests
	 * to auto-transition into capture mode, enable 'SAMPLING'.
	 * This will reduce the number of captures to the minimum needed
	 * to give the client a new image when requested.
	 *
	 * Note: 'SAMPLING' has no effect when in forced-capture mode.
	 */
	for (ci=0, sp = sample, sp_e = &sample[TotCam]; sp < sp_e; ci++, sp++) {
		/* if something sampled for this camera, recheck array */
		if (sp->tdiff != Cf->usertout * 8) {
			sp->samples[sp->sampidx++ % MAXTIMESAM] = sp->tdiff;
			if (Cf->sampling && CurState[ci] == act_autocap) {
				unsigned int minmsec;
				int si;

				sp->tdiff = Cf->usertout;
				for (si = 0; si < MAXTIMESAM; si++)
					if (sp->samples[si] < sp->tdiff)
						sp->tdiff = sp->samples[si];
				if ((minmsec = sp->tdiff * 1000) < Cf->capival)
					minmsec = Cf->capival;
				if (sp->curcapival != minmsec) {
					sp->curcapival = minmsec;
					capexec(cap_nsamp, ci, sp->curcapival);
				}
			}
		}

		/*
		 * Compare current state with number of clients and fill
		 * in the return vector appropriately.
		 */
		if (changes != NULL) {
			changes[ci] = act_none;
			if (CurState[ci] == act_down || CurState[ci] == act_cap)
				continue;	/* treat as immutable here */
			if (sp->clicnt > 0 && CurState[ci] != act_autocap)
				changes[ci] = act_autocap;
			else if (sp->clicnt == 0 && CurState[ci] != act_up)
				changes[ci] = act_up;
		}
	}

rsampout:
	sigprocmask(SIG_SETMASK, &osigmask, NULL);
	return totcli;
}

/*
 * parselog(log, cnt, changes) - parse apache log data into user/addr.
 *
 * This code expects to find: <ipaddr> - <user> [...] "GET ...<image>
 * Where:
 *	<ipaddr> begins with a digit,
 *	<user> is alpha-numeric or a single '-' (anonymous),
 *      <image> is the name of an image file.
 *
 * If matched, invoke climgr(), otherwise just ignore the line.
 *
 * Return number of successfully parsed log messages.
 */
int
parselog(char *log, int cnt, action_t *changes)
{
	char *ipaddr, *username, *slash, **clst;
	int newclient = 0;
	camidx_t ci;

	while (cnt > 0) {
		/* heuristics: if not on a newline, skip ahead */
		if (!isdigit(*log)) {
findnl:
			while (cnt > 0 && *log != '\n')
				cnt--, log++;
			if (cnt > 0)	/* skip past new line */
				++log, --cnt;
			continue;
		}

		/* required pattern: <digit><ipaddr> - <user> [...] "GET */
		ipaddr = log;
		do
			log++, cnt--;
		while (cnt > 0 && (isdigit(*log) || *log == '.'));

		/* " - <user>" */
		if (cnt < 4 || *log != ' ' || *(log + 1) != '-' ||
				*(log + 2) != ' ' ||
				!(isalnum(*(log + 3)) || *(log + 3) == '-'))
			goto findnl;

		*log = '\0';		/* terminate 'ipaddr' string */
		log += 3;		/* advance to user name */
		cnt -= 3;
		username = log;

		do
			log++, cnt--;
		while (cnt > 0 && isalnum(*log));
		if (cnt == 0 || *log != ' ')
			continue;
		*log = '\0';		/* terminate 'username' string */

		/* determine requested image and find associated camera */
		for ( ; cnt > 0 && *log != '\n'; cnt--, log++)
			if (*log == '"' && cnt > 3 && *(log + 1) == 'G' &&
			    *(log + 2) == 'E' && *(log + 3) == 'T')
				break;

		if (cnt == 0 || *log == '\n')	/* no '"GET', ignore it */
			continue;

		cnt -= 5;
		log += 5;
		slash = log;
		for ( ; cnt > 0 && *log != '\n'; cnt--, log++) {
			if (isspace(*log))
				break;
			if (*log == '/')
				slash = log;
		}

		if (cnt == 0 || *log == '\n' || slash == log)
			continue;		/* no file after GET */

		if (*slash == '/')
			slash++;
		*log = '\0';

		/*
		 * If a query string (question mark) is appended to
		 * the image file, terminate the URL there.  This
		 * allows clients to append a different (unique)
		 * string to each request to get around any caches.
		 */
		for (log = slash; *log != '\0'; log++)
			if (*log == '\?') {
				*log = '\0';
				break;
			}

		for (ci = 0; ci < TotCam; ci++)
			for (clst = Cf->camera[ci].imgfil;
					clst < Cf->camera[ci].imgfil_e; clst++)
				if (strcmp(*clst, slash) == 0)
					goto brkbrk;
brkbrk:
		if (ci == TotCam)		/* could not find camera */
			continue;

		/* add user to client connection directory */
		if (*username == '-')		/* unauthenticated client */
			username = "anonymous";
		climgr(cli_add, ci, username, ipaddr, changes);
		newclient++;
	}

	return newclient;
}

/*
 * camshow() - show status of server and all configured cameras.
 *
 * Server status is returned.
 */
action_t
camshow(const char *progname)
{
	const char *ofmt = "%s %s:%d [%s]\n";
	action_t actcur;
	pid_t pid;
	int i;

	sigprocmask(SIG_BLOCK, &sigmask_openpid, NULL);
	pid = readpid(pid_cur, NULL, NULL);
	sigprocmask(SIG_UNBLOCK, &sigmask_openpid, NULL);

	actcur = act_down;
	for (i = 0; i < TotCam; i++) {
		/* what is the "most active" state of all cameras? */
		switch(CurState[i]) {
		    case act_cap:		/* most active state */
			actcur = act_cap;
			break;
		    case act_autocap:		/* next most active state */
			if (actcur != act_cap)
				actcur = act_autocap;
			break;
		    case act_up:		/* least active state */
			if (actcur != act_cap && actcur != act_autocap)
				actcur = act_up;
		    default:
			break;
		}
	}
	printf(ofmt, "server", progname, pid, actmsg[actcur]);

	for (i = 0; i < TotCam; i++)
		printf(ofmt, "camera", Cf->camera[i].name, CapPid[i],
				actmsg[CurState[i]]);

	return actcur;
}

/*
 * clishow() - show connected clients by connection time, oldest first.
 *
 * This routine assumes the caller will soon exit because:
 *	o The working directory is changed to Cf->curcondir,
 *	o no attempt is made to free any allocated memory!
 *
 * Yes, we could easily do the above but *this* comment handles the
 * situation just as effectively and doesnt waste any cycles...
 */
void
clishow(void)
{
	DIR *dirp;
	struct dirent *dp;
	static struct dircli {
		struct dircli *next;
		char *camera, *name, *addr;
		time_t conntime;
	} *dchead = NULL, *dc, *dcold;

	if (Cf->curcondir == NULL || chdir(Cf->curcondir) < 0 ||
						(dirp = opendir(".")) == NULL)
		return;

	/* create a linked list of connection data starting at 'dchead' */
	while ((dp = readdir(dirp)) != NULL)
		if (*dp->d_name != '.') {
			struct stat sb;
			char *cp;

			if (stat(dp->d_name, &sb) < 0 ||
					(sb.st_mode & S_IFMT) != S_IFREG)
				continue;
			if ((dc = malloc(sizeof(struct dircli))) == NULL)
				break;
			if ((dc->camera = malloc(strlen(dp->d_name)+1)) == NULL)
				break;
			strcpy(dc->camera, dp->d_name);
			if ((cp = strchr(dc->camera, '_')) == NULL)
				continue;
			*cp++ = '\0';
			if ((cp = strchr(dc->name = cp, '_')) == NULL)
				continue;
			*cp++ = '\0';
			dc->addr = cp;
			dc->conntime = sb.st_mtime;
			dc->next = dchead;
			dchead = dc;
		}

	closedir(dirp);

	/* with each iteration over the list, print the oldest connection */
	for (;;) {
		struct tm *tm;
		time_t oldest = (time_t)0;

		for (dc = dchead; dc != NULL; dc = dc->next) {
			if (dc->conntime == 0)
				continue;
			if (oldest == 0 || dc->conntime < oldest) {
				oldest = dc->conntime;
				dcold = dc;
			}
		}

		if (oldest == (time_t)0)	/* none left */
			break;

		tm = localtime(&dcold->conntime);
		printf("client %s %s %s %02d/%02d/%d %02d:%02d:%02d\n",
			dcold->camera, dcold->name, dcold->addr,
			tm->tm_mon + 1, tm->tm_mday, 1900 + tm->tm_year,
			tm->tm_hour, tm->tm_min, tm->tm_sec);

		dcold->conntime = (time_t)0;	/* been there, done that */
	}
}

#define	PID_RDTOUT	60	/* secs to wait for server before killing */

static struct pidfil_s {
	char vers[4];			/* request/current data + version */
#define	PID_VERSCUR	"cV1\n"
#define	PID_VERSREQ	"rV1\n"
	pid_t pidsvr;			/* server pid */
	uid_t uidreq;			/* requestor's uid */
	char camreq[CONFCAMMAX];	/* requested camera states */
	char camcur[CONFCAMMAX];	/* current camera states */
	pid_t cappid[CONFCAMMAX];	/* current capture process leaders */
} pidfil;

static int fd_pid =		-1;

/*
 * openpid() - open PID file, return holding an exclusive lock.
 *
 * Call exit(2) on error (do not return).
 */
void
openpid(void)
{
	if (fd_pid != -1) {
		syslog(LOG_ERR, "open(%s): already open", Cf->pidfil);
		exit(2);
	}

	/* mode must be 600, otherwise anyone could open and lock us out */
	if ((fd_pid = open(Cf->pidfil, O_CREAT | O_RDWR, 0600)) < 0) {
		syslog(LOG_ERR, "open(%s): %m", Cf->pidfil);
		exit(2);
	}

	if (exlock(fd_pid, Cf->pidfil) != 0)	/* set exclusive lock on file */
		exit(2);
}

/*
 * readpid(typ, actcam, uid) - read PID file.
 *
 * If 'typ' is 'pid_req', the server process as asking for a client
 * request structure.  If no request pending, return 0.  Otherwise,
 * return server's (our) pid.
 *
 * If 'typ' is 'pid_cur', the client process will wait until the
 * server completes the state transition(s) before returning.  If
 * the server is down, return 0.  Otherwise, return server's pid.
 *
 * If pid file is not open, it is opened and closed before returning.
 * If pid file is already open, it will remain open on return.
 *
 * Return pid or 0 (depending on 'typ', see above).  Return requestors
 * uid in '*uid' (if non-NULL) and per-camera information in 'actcam[]'
 * (if non-NULL).  If 'typ' is 'pid_cur', global variables CurState[]
 * and CapPid[] are also filled in.
 *
 * Call exit(2) on error (do not return).
 */
pid_t
readpid(pidtyp_t typ, action_t *actcam, uid_t *uid)
{
	const char *noresp = "server not responding... killing in %d secs";
	ssize_t nr;
	int i, killed = 0, retry = PID_RDTOUT;
	pid_t mypid = getpid();

again:
	/*
	 * If pid file is not already opened, open it before doing
	 * the lseek/read of the 'pidfil_s' struct.  If we had to
	 * open it, then immediately close it again after the read.
	 * Exit on any error.
	 */
	{
		int opened = 0;

		if (fd_pid == -1) {
			openpid();
			opened = 1;
		}
		if (lseek(fd_pid, (off_t)0, SEEK_SET) < 0) {
			syslog(LOG_ERR, "lseek(%s): %m", Cf->pidfil);
			exit(2);
		}
		if ((nr = read(fd_pid, &pidfil, sizeof pidfil)) < 0) {
			syslog(LOG_ERR, "read(%s): %m", Cf->pidfil);
			exit(2);
		}
		if (opened)
			closepid();
	}

	/*
	 * Check result of the read.  If requesting current status and
	 * the file was empty, assume the pid file was just created
	 * (and the server is down).  If the read was short, just exit.
	 */
	if (typ == pid_cur && nr == 0) {
		goto svrdown;
	} else if (nr != sizeof(pidfil)) {
		syslog(LOG_ERR, "read(%s): incomplete read", Cf->pidfil);
		exit(2);
	}

	/*
	 * Do some validity checks on the structure.  Make sure the
	 * version is okay and ensure that all actions are valid.
	 */
	if (strncmp(pidfil.vers, PID_VERSCUR, sizeof pidfil.vers) != 0 &&
	    strncmp(pidfil.vers, PID_VERSREQ, sizeof pidfil.vers) != 0) {
		syslog(LOG_ERR, "read(%s): unsupported version",
			Cf->pidfil);
		exit(2);
	}
	for (i = 0; i < TotCam; i++) {
		if (((*pidfil.vers == 'r' &&
				(action_t)pidfil.camreq[i] < act_none) ||
			(action_t)pidfil.camreq[i] > act_down) ||
		    ((*pidfil.vers == 'c' &&
				(action_t)pidfil.camcur[i] < act_none) ||
			(action_t)pidfil.camcur[i] > act_down))
			break;
	}
	if (pidfil.pidsvr == 0 || i < TotCam) {
		syslog(LOG_ERR, "read(%s): corrupt file (%s)", Cf->pidfil,
			pidfil.pidsvr == 0? "server pid 0": "invalid action");
		exit(2);
	}

	switch (typ) {
	    case pid_cur:		/* client request */
		/*
		 * Determine if 'pidfil.pidsvr' is around and owned by us.
		 * If this pid was reused by another uid, the kill() will
		 * fail.  There's a slight chance that this pid is owned by
		 * us but is not a 'cammgr' server (e.g., another $CAMOWN
		 * process), making the point that $CAMOWN should be a
		 * unique user (i.e., not "nobody", definitely not "root").
		 * Then again, it's probably not worth worrying about.
		 */
		if (kill(pidfil.pidsvr, 0) < 0)
			goto svrdown;

		/*
		 * If a request is still pending, close pid file to give
		 * server a chance to field it.  Sleep a bit, reopen/retry.
		 */
		if (*pidfil.vers == 'r') {
			if (retry != PID_RDTOUT && (retry % 10) == 0) {
				syslog(LOG_WARNING, noresp, retry);
				/* doesnt hurt, but probably wont help either */
				(void) kill(pidfil.pidsvr, SIGUSR1);
			}

			if (fd_pid == -1) {
				sleep(1);
			} else {
				closepid();
				sleep(1);
				openpid();
			}

			if (retry-- > 0)
				goto again;
			/*
			 * Never saw this happen, but just in case, we
			 * cant leave this setuid process hanging around.
			 */
			if (killed == 0) {
				syslog(LOG_ERR,
				   "killing unresponsive server (pid %d)...",
				   pidfil.pidsvr);
				kill(pidfil.pidsvr, SIGTERM);
				sleep(3);
				kill(pidfil.pidsvr, SIGKILL);
			}
			if (++killed < 3)	/* try a few more times */
				goto again;
			exit(2);
		}

		/* fill in CurState[], CapPid[], actcam[], and *uid */
		for (i = 0; i < CONFCAMMAX; i++)
			CurState[i] = (action_t)pidfil.camcur[i];
		bcopy(pidfil.cappid, (void *)CapPid, sizeof CapPid);
		if (actcam != NULL)
			bcopy((action_t *)CurState, actcam, sizeof actcam);
		if (uid != NULL)
			*uid = pidfil.uidreq;
		break;

	    case pid_req:		/* server request */
		/*
		 * The server is looking for a request.  First do a sanity
		 * check to ensure we are the server, then check that a
		 * request is pending (i.e., that we didnt miss it).
		 */
		if (pidfil.pidsvr != mypid) {
			syslog(LOG_ERR, "pid mismatch: %d != %d, exiting...",
				pidfil.pidsvr, mypid);
			exit(2);
		}
		if (*pidfil.vers == 'c')
			return 0;

		/* fill in actcam[] and *uid */
		if (actcam != NULL)
			for (i = 0; i < CONFCAMMAX; i++)
				actcam[i] = (action_t)pidfil.camreq[i];
		if (uid != NULL)
			*uid = pidfil.uidreq;
		break;

	    case pid_curonly:		/* impossible */
		syslog(LOG_ERR, "readpid: invalid argument: pid_curonly");
		exit(2);
	}

	return pidfil.pidsvr;

svrdown:
	/*
	 * If we ended up here, the request was from a client looking for
	 * current status ('pid_cur'), but the server is down.  Fill in
	 * CurState[], CapPid[], actcam[], and *uid accordingly.
	 */
	for (i = 0; i < CONFCAMMAX; i++)
		CurState[i] = act_down;
	bzero((void *)CapPid, sizeof CapPid);
	if (actcam != NULL)
		bcopy((action_t *)CurState, actcam, sizeof actcam);
	if (uid != NULL)
		*uid = MyUid;

	return 0;
}

/*
 * writepid(typ, actcam, pid) - write args to opened PID file.
 *
 * Fill in a 'pidfil_s' structure and write it out.
 *
 * Caller must have previously opened pid file.
 *
 * Call exit(2) on error (do not return).
 */
void
writepid(pidtyp_t typ, const action_t *actcam, pid_t pid)
{
	ssize_t nr;
	int i;

	switch (typ) {
	    case pid_cur:		/* server writing current status */
		bcopy(PID_VERSCUR, pidfil.vers, sizeof pidfil.vers);
		pidfil.pidsvr = pid;
		/*FALLTHRU*/

	    case pid_curonly:		/* write current status only */
		for (i = 0; i < CONFCAMMAX; i++)
			pidfil.camcur[i] = CurState[i];
		bcopy((void *)CapPid, pidfil.cappid, sizeof pidfil.cappid);
		break;

	    case pid_req:		/* client writing server request */
		bcopy(PID_VERSREQ, pidfil.vers, sizeof pidfil.vers);
		for (i = 0; i < CONFCAMMAX; i++)
			pidfil.camreq[i] = actcam[i];
		pidfil.uidreq = MyUid;
		break;
	}

	if (lseek(fd_pid, (off_t)0, SEEK_SET) < 0) {
		syslog(LOG_ERR, "lseek(%s): %m", Cf->pidfil);
		exit(2);
	}
	if ((nr = write(fd_pid, &pidfil, sizeof pidfil)) < 0) {
		syslog(LOG_ERR, "write(%s): %m", Cf->pidfil);
		exit(2);
	}
	if (nr != sizeof pidfil) {
		syslog(LOG_ERR, "write(%s): incomplete write", Cf->pidfil);
		exit(2);
	}
}

/*
 * closepid() - close PID file (if open).
 */
void
closepid(void)
{
	if (fd_pid >= 0) {
		(void) close(fd_pid);
		fd_pid = -1;
	}
}
